from unittest import TestCase
from croydon.project import parse_field_descriptor, normalize_model_name


class TestProject(TestCase):
    def test_parse_invalid_subtyped(self):
        with self.assertRaises(ValueError):
            parse_field_descriptor("name:str[subtype]")

    def test_parse_non_subtyped_ref(self):
        with self.assertRaises(ValueError):
            parse_field_descriptor("user_id:ref")

    def test_parse_no_type(self):
        fd = parse_field_descriptor("name")
        self.assertEqual(fd.name, "name")
        self.assertEqual(fd.field_type, "StringField")
        self.assertEqual(fd.field_subtype, None)

    def test_various_types(self):
        fd = parse_field_descriptor("count:int")
        self.assertEqual(fd.name, "count")
        self.assertEqual(fd.field_type, "IntField")
        self.assertEqual(fd.field_subtype, None)
        self.assertFalse(fd.required)

        fd = parse_field_descriptor("user_id:ref[User]!")
        self.assertEqual(fd.name, "user_id")
        self.assertEqual(fd.field_type, "ReferenceField")
        self.assertEqual(fd.field_subtype, "User")
        self.assertTrue(fd.required)

    def test_render(self):
        fd = parse_field_descriptor("user_id:ref[User]!")
        self.assertEqual(
            "user_id: ReferenceField[User] = ReferenceField(reference_model=User, required=True)",
            fd.render(),
        )

    def test_normalize(self):
        self.assertEqual("UserDef", normalize_model_name("user_def"))
        self.assertEqual("UserDef", normalize_model_name("UserDef"))
        self.assertEqual("UserDef", normalize_model_name("User_Def"))



from unittest import TestCase
from croydon.config import BaseConfig


class TestConfig(TestCase):

    def test_config(self):
        cfg = BaseConfig()
        self.assertEqual(cfg.get("session.cookie"), cfg.session.cookie)
        with self.assertRaises(KeyError):
            cfg.get("some.incorrect.key")



import re
import os
import inspect
from importlib import import_module
from string import ascii_uppercase
from typing import Optional, Literal, Dict, List
from pydantic import BaseModel


FD_EXPR = re.compile(r"^(\w+)(:(\w+)(\[([^\]]+)\])?)?(!?)$")
SUBTYPED_TYPES = {"ListField", "DictField", "ReferenceField"}
FieldType = Literal[
    "StringField",
    "IntField",
    "FloatField",
    "BoolField",
    "ListField",
    "DictField",
    "ReferenceField",
    "SelfReferenceField",
    "ObjectIdField",
    "DatetimeField",
]


class FieldDescriptor(BaseModel):
    name: str
    field_type: FieldType
    field_subtype: Optional[str] = None
    required: bool = False

    def render(self) -> str:
        args = {}
        decl = self.name
        if self.field_subtype:
            decl += f": {self.field_type}[{self.field_subtype}]"
        decl += f" = {self.field_type}("

        if self.field_type == "ReferenceField":
            args["reference_model"] = self.field_subtype
        if self.required:
            args["required"] = "True"

        if args:
            decl += ", ".join([f"{k}={v}" for k, v in args.items()])

        decl += ")"
        return decl


def parse_field_descriptor(fd: str) -> FieldDescriptor:
    match = FD_EXPR.findall(fd)
    if not match:
        raise ValueError(f'invalid field descriptor "{fd}"')

    tokens = match[0]

    fname, _, ftype, _, fsubtype, excl = tokens
    match ftype.lower():
        case "":
            field_type = "StringField"
        case "str" | "string" | "stringfield":
            field_type = "StringField"
        case "int" | "integer" | "integerfield":
            field_type = "IntField"
        case "bool" | "boolean" | "boolfield":
            field_type = "BoolField"
        case "float" | "floatfield":
            field_type = "FloatField"
        case "objid" | "objectid" | "objectidfield":
            field_type = "ObjectIdField"
        case "list" | "listfield":
            field_type = "ListField"
        case "dict" | "dictfield":
            field_type = "DictField"
        case "date" | "datetime" | "timestamp" | "ts" | "dt":
            field_type = "DatetimeField"
        case "ref" | "reference" | "referencefield":
            field_type = "ReferenceField"
        case "self" | "selfreference" | "selfreferencefield":
            field_type = "SelfReferenceField"
        case _:
            raise ValueError(f'unknown field type "{ftype}"')

    if fsubtype and field_type not in SUBTYPED_TYPES:
        raise ValueError(f"type {field_type} can't have subtypes")
    if not fsubtype and field_type == "ReferenceField":
        raise ValueError(f"type ReferenceField must have a subtype")

    required = False
    if excl:
        required = True

    return FieldDescriptor(
        name=fname,
        field_type=field_type,
        field_subtype=fsubtype or None,
        required=required,
    )


def has_caps(token: str) -> bool:
    for sym in token:
        if sym in ascii_uppercase:
            return True
    return False


def normalize_model_name(token: str) -> str:
    if not has_caps(token):
        token = token.title()
    return token.replace("_", "")


def find_models(app_models_path: str) -> Dict[str, str]:
    from croydon.models import StorableModel

    model_imports = {}
    for _, _, files in os.walk(app_models_path):
        for filename in files:
            if filename.endswith(".py"):
                modname = "app.models." + filename[:-3].replace("/", ".")
                if modname.endswith(".__init__"):
                    modname = modname[:-9]
                module = import_module(modname)
                for item_name in dir(module):
                    item = getattr(module, item_name)
                    if inspect.isclass(item) and issubclass(item, StorableModel):
                        if item.__name__ not in model_imports:
                            model_imports[item.__name__] = item.__module__
    return model_imports



import enum
import pymongo.collation

from typing import Iterable, NamedTuple, Type, TypeVar, TypedDict


class IndexDirection(int, enum.Enum):
    ASCENDING = pymongo.ASCENDING
    DESCENDING = pymongo.DESCENDING

    def __str__(self) -> str:
        """This is necessary to unbreak pymongo.helpers._gen_index_name"""
        return str(self.value)


class IndexType(str, enum.Enum):
    GEO2D = pymongo.GEO2D
    GEOSPHERE = pymongo.GEOSPHERE
    HASHED = pymongo.HASHED
    TEXT = pymongo.TEXT

    def __str__(self) -> str:
        """This is necessary to unbreak pymongo.helpers._gen_index_name"""
        return str(self.value)


IndexSpec = IndexDirection | IndexType


class IndexKey(NamedTuple):
    key: str
    spec: IndexSpec


class IndexOptions(TypedDict, total=False):
    """NB: This is not a full list of possible options
    See https://www.mongodb.com/docs/manual/reference/command/createIndexes/
    """
    name: str
    unique: bool
    background: bool
    sparse: bool
    bucketSize: int
    bits: int
    min: float
    max: float
    expireAfterSeconds: int
    collation: pymongo.collation.Collation


class _Index(NamedTuple):
    keys: Iterable[IndexKey]
    options: IndexOptions


TIndex = TypeVar("TIndex", bound=_Index)


# This is to apply default value generator, impossible directly in NamedTuple
class Index(_Index):
    def __new__(
        cls: Type[TIndex],
        keys: Iterable[IndexKey],
        options: IndexOptions | None = None
    ) -> TIndex:
        if options is None:
            options = IndexOptions()
        return super().__new__(cls, keys, options)



from typing import List
from uuid import uuid4
from datetime import datetime
from bson.objectid import ObjectId, InvalidId


def uuid4_string() -> str:
    return str(uuid4())


def resolve_id(id_: str | ObjectId | None) -> ObjectId | str | None:
    # ObjectId(None) generates a new unique object id
    # We need to handle that case and return None instead
    if id_ is not None:
        if not isinstance(id_, ObjectId):
            try:
                objid_expr = ObjectId(id_)
                if str(objid_expr) == id_:
                    return objid_expr
            except (InvalidId, TypeError):
                pass
    return id_


# Mongo stores datetime rounded to milliseconds as its datetime
# capabilities are limited by v8 engine
def now() -> datetime:
    dt = datetime.utcnow()
    dt = dt.replace(microsecond=dt.microsecond // 1000 * 1000)
    return dt


NilObjectId: ObjectId = ObjectId("000000000000000000000000")


class ObjectIdParam(str):

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v: str) -> ObjectId:
        try:
            return ObjectId(v)
        except InvalidId:
            raise ValueError("invalid object id")


class ObjectIdListParam(str):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v: str) -> List[ObjectId]:
        try:
            return [ObjectId(token.strip()) for token in v.split(",")]
        except InvalidId:
            raise ValueError("invalid object id list")


class StringListParam(str):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v: str) -> List[str]:
        return [token.strip() for token in v.split(",")]



from typing import TypeVar
from .abc import AbstractCache
from .no_cache import NoCache
from .trace_cache import TraceCache
from .simple import SimpleCache
from .memcached import MemcachedCache
from .req_local_cache import RequestLocalCache

TCache = TypeVar("TCache", bound=AbstractCache)



from typing import Optional, Callable, Awaitable
from fastapi import Header, Depends, Request

from croydon import ctx
from croydon.errors import AuthenticationError

from app.models import Token, User, Session


async def auth_token(authorization: Optional[str] = Header(default=None)) -> Optional[Token]:
    if authorization is None:
        return None

    parts = authorization.split(" ")
    if len(parts) != 2:
        return None

    token_type, token = parts

    if token_type != "Token":
        return None

    token = await Token.cache_get(token)
    if token:
        if token.is_expired():
            ctx.log.debug(f"token {token.token} is expired")
            return None
        await token.auto_prolong()
    return token


def authenticated_user(required: bool = True) -> Callable[[Optional[Token]], Awaitable[Optional[User]]]:
    async def inner(token: Optional[Token] = Depends(auth_token),
                    session: Optional[Session] = Depends(current_session)) -> Optional[User]:
        if token is not None:
            user = await token.user()
            if user:
                ctx.log.debug("authenticated_user found by token")
                return user

        if session is not None:
            user = await session.user()
            if user:
                ctx.log.debug("authenticated_user found by session")
                return user

        if required:
            ctx.log.debug("authenticated_user was not found but auth is required")
            raise AuthenticationError()

        ctx.log.debug("authenticated_user was not found")
        return None

    return inner


async def current_session(request: Request) -> Session:
    return request.state.session





from typing import (
    Generic, Callable, Optional, List, Type,
    AsyncIterator, Any, Dict, Sequence, Union, Mapping, Tuple
)
from time import sleep
from datetime import datetime

import pymongo
from bson.objectid import ObjectId, InvalidId
from motor.motor_asyncio import AsyncIOMotorClient
from motor.motor_asyncio import AsyncIOMotorCursor
from motor.motor_asyncio import AsyncIOMotorDatabase
from mongomock_motor import (
    AsyncMongoMockClient, AsyncMongoMockDatabase, AsyncCursor
)
from pymongo import ReturnDocument
from pymongo.results import DeleteResult, UpdateResult
from pymongo.errors import ServerSelectionTimeoutError
from .config import DatabaseConfig
from .types import RT, TModel, KwArgs, QueryExpression, QueryType, CommonDict

MONGO_RETRY_MAX: int = 6
MONGO_RETRY_SLEEP: int = 3


def intercept_pymongo_errors(func: Callable[..., RT]) -> Callable[..., RT]:
    def wrapper(self: "Shard", *args: List[Any], **kwargs: KwArgs) -> RT:
        retries = MONGO_RETRY_MAX
        while True:
            retries -= 1
            try:
                return func(self, *args, **kwargs)
            except ServerSelectionTimeoutError:
                from .context import ctx
                ctx.log.error(
                    "ServerSelectionTimeoutError in db module, retries left = %d",
                    retries,
                )
                if retries == MONGO_RETRY_MAX // 2:
                    ctx.log.error("trying to reinstantiate connection")
                    db_inst = self
                    db_inst.reset_conn()
                elif retries == 0:
                    ctx.log.error(
                        "ServerSelectionTimeoutError max retries exceeded, giving up"
                    )
                    raise
                sleep(MONGO_RETRY_SLEEP)

    return wrapper


class ObjectsCursor(Generic[TModel], AsyncIterator[TModel]):
    cursor: Union[AsyncIOMotorCursor, AsyncCursor]

    def __init__(self, cursor: Union[AsyncIOMotorCursor, AsyncCursor], ctor: Type[TModel], query: QueryType,
                 shard_id: Optional[str] = None) -> None:
        self.ctor = ctor
        self.cursor = cursor
        self.stored_query = query
        self._shard_id = shard_id

    async def all(self, max_length: Optional[int] = None) -> List[TModel]:
        return [self.ctor(x) for x in await self.cursor.to_list(max_length)]

    async def as_map(self, max_length: Optional[int] = None) -> Dict[ObjectId, TModel]:
        items = await self.all(max_length)
        item_map = {}
        for item in items:
            item_map[item.id] = item
        return item_map

    async def all_dict(self,
                       max_length: Optional[int] = None,
                       fields: Optional[List[str]] = None) -> List[Dict[str, Any]]:
        return [self.ctor(x).to_dict(fields=fields) for x in await self.cursor.to_list(max_length)]

    def limit(self, limit: int) -> "ObjectsCursor[TModel]":
        self.cursor.limit(limit)
        return self

    def skip(self, skip: int) -> "ObjectsCursor[TModel]":
        self.cursor.skip(skip)
        return self

    def sort(self, key_or_list: Union[str, Sequence[Tuple[str, Union[int, str, Mapping[str, Any]]]]],
             direction: Optional[Union[int, str]] = None) -> "ObjectsCursor[TModel]":
        if isinstance(key_or_list, str) and direction is None:
            direction = pymongo.ASCENDING
        self.cursor.sort(key_or_list, direction)
        return self

    async def count(self) -> int:
        # usually self.ctor contains a classmethod `ctor` of a StorableModel
        # hence this hack
        ctor = self.ctor.__self__
        return await ctor.count(self.stored_query)

    def __aiter__(self) -> AsyncIterator["TModel"]:
        return self

    async def __anext__(self) -> "TModel":
        item = await self.cursor.__anext__()
        if self._shard_id:
            return self.ctor(shard_id=self._shard_id, attrs=item)
        else:
            return self.ctor(attrs=item)

    def __getattr__(self, item: str) -> Any:
        return getattr(self.cursor, item)


class Shard:
    _config: DatabaseConfig
    _client: Optional[Union[AsyncIOMotorClient, AsyncMongoMockClient]]
    _conn: Optional[Union[AsyncIOMotorDatabase, AsyncMongoMockDatabase]]
    _mock: bool

    def __init__(self, config: DatabaseConfig, mock: bool = False) -> None:
        self._config = config
        self._client = None
        self._conn = None
        self._mock = mock

    def get_client(self) -> Union[AsyncIOMotorClient, AsyncMongoMockClient]:
        if not self._client:
            if self._mock:
                self._client = AsyncMongoMockClient(self._config.uri)
            else:
                self._client = AsyncIOMotorClient(self._config.uri)
        return self._client

    def reset_conn(self) -> None:
        self._client = None
        self._conn = None

    def init_conn(self) -> None:
        from .context import ctx
        if not self._mock:
            ctx.log.info("creating a new mongo connection")
        client = self.get_client()
        self._conn = client.get_database()

    @property
    def conn(self) -> Union[AsyncIOMotorDatabase, AsyncMongoMockDatabase]:
        if self._conn is None:
            self.init_conn()
        return self._conn

    @intercept_pymongo_errors
    async def get_obj(self,
                      ctor: Callable[..., "TModel"],
                      collection: str,
                      query: QueryExpression) -> Optional["TModel"]:
        if not isinstance(query, dict):
            try:
                query = {"_id": ObjectId(query)}
            except InvalidId:
                pass
        data = await self.conn[collection].find_one(query)
        if data:
            return ctor(attrs=data)

        return None

    @intercept_pymongo_errors
    async def get_obj_id(self, collection: str, query: QueryType) -> Optional[ObjectId]:
        obj = await self.conn[collection].find_one(
            query, projection=()
        )
        if obj:
            return obj["_id"]
        return None

    @intercept_pymongo_errors
    def get_objs(self,
                 ctor: Callable[..., "TModel"],
                 collection: str,
                 query: QueryType,
                 **kwargs: Dict[str, Any]) -> ObjectsCursor[TModel]:
        cursor = self.conn[collection].find(query, **kwargs)
        return ObjectsCursor(cursor, ctor, query)

    @intercept_pymongo_errors
    def get_objs_projected(self,
                           collection: str,
                           query: CommonDict,
                           projection: CommonDict, **kwargs: KwArgs) -> AsyncIOMotorCursor:
        cursor = self.conn[collection].find(query, projection=projection, **kwargs)
        return cursor

    @intercept_pymongo_errors
    def get_aggregated(self, collection: str, pipeline: List[CommonDict], **kwargs: KwArgs) -> AsyncIOMotorCursor:
        cursor = self.conn[collection].aggregate(pipeline, **kwargs)
        return cursor

    @intercept_pymongo_errors
    async def count_docs(self, collection: str, query: CommonDict, **kwargs: KwArgs) -> int:
        coll = self.conn[collection]
        if query == {}:
            return await coll.estimated_document_count(**kwargs)
        else:
            return await coll.count_documents(query, **kwargs)

    def get_objs_by_field_in(self,
                             ctor: Type[TModel],
                             collection: str,
                             field: str,
                             values: List[str],
                             **kwargs: KwArgs) -> ObjectsCursor[TModel]:
        return self.get_objs(ctor, collection, {field: {"$in": values}}, **kwargs)

    @intercept_pymongo_errors
    async def save_obj(self, obj: "TModel") -> None:
        if obj.is_new:
            # object to_dict() method should always return all fields
            data = obj.to_dict(include_restricted=True, convert_id=True)
            # although with the new object we shouldn"t pass _id=null to mongo
            del data["_id"]
            insert_result = await self.conn[obj.collection].insert_one(data)
            obj.id = insert_result.inserted_id
        else:
            await self.conn[obj.collection].replace_one(
                {"_id": obj.id},
                obj.to_dict(include_restricted=True, convert_id=True),
                upsert=True,
            )

    @intercept_pymongo_errors
    async def delete_obj(self, obj: "TModel") -> None:
        if obj.is_new:
            return
        await self.conn[obj.collection].delete_one({"_id": obj.id})

    @intercept_pymongo_errors
    async def find_and_update_obj(self, obj: "TModel", update: CommonDict, when: Optional[CommonDict] = None) -> None:
        query = {"_id": obj.id}
        if when:
            assert "_id" not in when
            query.update(when)

        new_data = await self.conn[obj.collection].find_one_and_update(
            query, update, return_document=ReturnDocument.AFTER
        )
        return new_data

    @intercept_pymongo_errors
    async def delete_query(self, collection: str, query: CommonDict) -> DeleteResult:
        return await self.conn[collection].delete_many(query)

    @intercept_pymongo_errors
    async def update_query(self, collection: str, query: CommonDict, update: CommonDict) -> UpdateResult:
        return await self.conn[collection].update_many(query, update)

    # SESSIONS
    @intercept_pymongo_errors
    async def get_session(self, sid: str, collection: str = "sessions") -> CommonDict:
        return await self.conn[collection].find_one({"sid": sid})

    @intercept_pymongo_errors
    async def update_session(self, sid: str, data: CommonDict, expiration: datetime,
                             collection: str = "sessions") -> None:
        await self.conn[collection].update_one(
            {"sid": sid}, {"$set": {"sid": sid, "data": data, "expiration": expiration}}, True
        )

    @intercept_pymongo_errors
    async def cleanup_sessions(self, collection: str = "sessions") -> int:
        result = await self.conn[collection].delete_one({"expiration": {"$lt": datetime.now()}})
        return result.deleted_count


class DB:
    MONGODB_INFO_FIELDS = (
        "allocator",
        "bits",
        "debug",
        "gitVersion",
        "javascriptEngine",
        "maxBsonObjectSize",
        "modules",
        "ok",
        "openssl",
        "storageEngines",
        "sysInfo",
        "version",
        "versionArray",
    )

    __meta: Optional[Shard]

    def __init__(self,
                 *,
                 config: DatabaseConfig,
                 mock: bool = False) -> None:
        self.reconfigure(config, mock)

    def reconfigure(self, config: DatabaseConfig, mock: bool = False) -> None:
        self.__meta = Shard(config, mock=mock)

    @property
    def meta(self) -> Shard:
        return self.__meta



import os.path
import inspect
import pickle
from typing import Callable, Awaitable, Any, List, Optional
from functools import wraps
from time import time
from datetime import timedelta

from .types import KwArgs, TModel, TModelType, RT
from .errors import ObjectSaveRequired


def save_required(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    save_required decorator is meant to decorate model methods that
    require a model instance to be saved first.

    For example, StorableModel's own db_update method makes no sense
    until the instance is persisted in the database
    """
    @wraps(func)
    def wrapper(self, *args: List[Any], **kwargs: KwArgs) -> Any:
        if self.is_new:
            raise ObjectSaveRequired("This object must be saved first")
        return func(self, *args, **kwargs)

    return wrapper


class api_field:
    """
    api_field is a decorator for model no-args methods (including async),
    allowing API users to access certain methods, i.e. implementing computed fields
    """

    def __init__(self, fn):
        if not inspect.isfunction(fn):
            raise RuntimeError("only functions and methods can be decorated with api_field")
        self.fn = fn

    def __set_name__(self, owner, name):
        from .models.fields import ComputedField
        fd = ComputedField(is_async=inspect.iscoroutinefunction(self.fn))

        # computed_fields are inherited from MetaModel down the class tree
        # here we make sure every descendant does not modify parents' computed fields

        computed_fields = owner._computed_fields.copy()
        for base in owner.__bases__:
            if hasattr(base, "_computed_fields"):
                computed_fields.update(base._computed_fields)
        computed_fields[name] = fd
        owner._computed_fields = computed_fields
        owner._computed_fields[name] = fd
        setattr(owner, name, self.fn)


class model_cached_method:
    """
    model_cached_method decorates model methods. The wrapper provided
    generates a unique cache key out of
    - model class name
    - model id
    - method name

    e.g. User.6433d2bfd5a353b0a6822d7f.full_name

    and tries to fetch the value from cache. If value is not there, the
    original method is called and the value is stored to cache before
    being returned

    The methods wrapped with this decorator are automatically added to a
    class property called _cached_methods: set. This allows StorableModel.invalidate
    method to invalidate these cached values along with other model-provided cache.
    """

    orig_func: Callable[[TModel, ...], Awaitable[RT]] = None

    def __init__(self, func: Callable[[TModel, ...], Awaitable[RT]]) -> None:
        self.orig_func = func

    def __set_name__(self, owner: TModelType, name: str) -> None:

        @wraps(self.orig_func)
        async def wrapper(this: TModel, *args, **kwargs) -> RT:
            from .context import ctx
            cache_key = f"{this.collection}.{this.id}.{self.orig_func.__name__}"

            t1 = time()
            value = await ctx.cache_l1.get(cache_key)
            if value is not None:
                td = time() - t1
                ctx.log.debug(
                    "%s L1 hit %s %.3f secs", ctx.cache_l1.NAME, cache_key, td
                )
                return value

            t1 = time()
            value = await ctx.cache_l2.get(cache_key)
            if value is not None:
                td = time() - t1
                ctx.log.debug(
                    "%s L2 hit %s %.3f secs", ctx.cache_l2.NAME, cache_key, td
                )
                await ctx.cache_l1.set(cache_key, value)
                return value

            t1 = time()
            value = await self.orig_func(this, *args, **kwargs)
            td = time() - t1
            ctx.log.debug("%s miss %s %.3f secs", ctx.cache_l2.NAME, cache_key, td)
            await ctx.cache_l2.set(cache_key, value)
            await ctx.cache_l1.set(cache_key, value)

            return value

        owner._cached_methods.add(name)
        setattr(owner, name, wrapper)


def config_file_cache(cfg_key: str, *, ttl: Optional[timedelta] = None) -> Callable[[Callable[..., RT]], Callable[..., RT]]:
    """
    config_file_cache decorates a function and caches its output to disk

    The filename is taken from croydon context, namely from your app configuration by a given key.
    Optional ttl argument is a timedelta within which the cached data is considered valid. After expiration
    the file is overwritten with a new function output

    The function _does not_ take arguments into account, i.e. there's only one cache per function so it's
    only suitable for decorating arg-less functions
    """
    def decorator(func: Callable[..., RT]) -> Callable[..., RT]:

        @wraps(func)
        async def async_wrapper(*args, **kwargs) -> RT:
            from croydon import ctx

            try:
                filename = ctx.cfg.get(cfg_key)
            except KeyError:
                ctx.log.error(f"error reading config key {cfg_key}, bypassing filecache")
                return await func(*args, **kwargs)

            if os.path.isfile(filename):
                expired = False
                if ttl is not None:
                    st = os.stat(filename)
                    age = time() - st.st_mtime
                    if ttl.total_seconds() < age:
                        expired = True
                if not expired:
                    try:
                        with open(filename, "rb") as cf:
                            return pickle.load(cf)
                    except (EnvironmentError, pickle.PickleError):
                        pass

            value = await func(*args, **kwargs)
            with open(filename, "wb") as cf:
                pickle.dump(value, cf)
            return value

        @wraps(func)
        def sync_wrapper(*args, **kwargs) -> RT:
            from croydon import ctx

            try:
                filename = ctx.cfg.get(cfg_key)
            except KeyError:
                ctx.log.error(f"error reading config key {cfg_key}, bypassing filecache")
                return func(*args, **kwargs)

            if os.path.isfile(filename):
                expired = False
                if ttl is not None:
                    st = os.stat(filename)
                    age = time() - st.st_mtime
                    if ttl.total_seconds() < age:
                        expired = True
                if not expired:
                    try:
                        with open(filename, "rb") as cf:
                            return pickle.load(cf)
                    except (EnvironmentError, pickle.PickleError):
                        pass

            value = func(*args, **kwargs)
            with open(filename, "wb") as cf:
                pickle.dump(value, cf)
            return value

        return async_wrapper if inspect.iscoroutinefunction(func) else sync_wrapper

    return decorator



import re
import inspect
from abc import ABC
from typing import Callable, Generic, Hashable, Protocol, Tuple, TypeVar, Type, Iterable, overload
from numbers import Real
from bson import ObjectId
from datetime import datetime
from dataclasses import dataclass
from ..types import TModelType
from ..errors import ValidationError
from .index import IndexSpec, IndexOptions, IndexDirection
from .reference import OnDestroy, ModelReference

T = TypeVar("T")

TBaseFieldDescriptor = TypeVar(
    "TBaseFieldDescriptor",
    bound="BaseFieldDescriptor",
)


class BaseFieldDescriptor(Generic[T]):
    name: str = ""

    def __set__(self, obj: object, value: T | None) -> None:
        obj.__dict__[self.name] = value

    def __set_name__(self, owner: Type[object], name: str) -> None:
        self.name = name

    @overload
    def __get__(
        self: "TBaseFieldDescriptor[T]",
        obj: None,
        objtype: Type[object] | None = None
    ) -> "TBaseFieldDescriptor[T]": ...

    @overload
    def __get__(
        self: "TBaseFieldDescriptor[T]",
        obj: object,
        objtype: Type[object] | None = None
    ) -> T | None: ...

    def __get__(
        self: "TBaseFieldDescriptor[T]",
        obj: object | None,
        objtype: Type[object] | None = None
    ) -> T | None | "TBaseFieldDescriptor[T]":
        if obj is None:
            return self
        return obj.__dict__[self.name]


class FieldProto(Protocol):
    rejected: bool
    restricted: bool
    index: IndexSpec | None
    index_options: IndexOptions

    def set_default(self, obj: object) -> None: ...
    async def validate(self, obj: object) -> None: ...


IndexInput = bool | IndexSpec | None


class Field(BaseFieldDescriptor[T]):

    required: bool
    rejected: bool
    restricted: bool
    index: IndexSpec | None
    index_options: IndexOptions
    choices: set[T] | None
    def_value: T | Callable[[], T] | None

    __explicit_types__: Iterable[Type] = []

    def __init__(
        self,
        *,
        required: bool = False,
        rejected: bool = False,
        restricted: bool = False,
        default: T | Callable[[], T] | None = None,
        index: IndexInput | None = None,
        unique: bool = False,
        choices: Iterable[T] | None = None,
    ) -> None:
        self.required = required
        self.rejected = rejected
        self.restricted = restricted
        if choices:
            self.choices = set(choices)
        else:
            self.choices = None
        self.def_value = default
        self.index, self.index_options = self.__generate_index(index, unique)

    def set_default(self, obj: object) -> None:
        value = self.def_value
        if callable(value):
            value = value()
        setattr(obj, self.name, value)

    @staticmethod
    def __generate_index(
        index: IndexInput, unique: bool
    ) -> Tuple[IndexSpec | None, IndexOptions]:
        options = IndexOptions()
        if unique:
            if not isinstance(index, IndexSpec):
                index = IndexDirection.ASCENDING
            options["unique"] = True

        if not isinstance(index, IndexDirection):
            if index:
                index = IndexDirection.ASCENDING
            else:
                index = None

        return index, options

    def set_name(self, name: str) -> None:
        self.name = name

    async def validate(self, obj: object) -> None:
        value = self.__get__(obj)
        if self.required and value is None:
            raise ValidationError(f"field {self.name} is required")

        if value is not None and self.__explicit_types__:
            if not any([isinstance(value, _type) for _type in self.__explicit_types__]):
                type_names = [x.__name__ for x in self.__explicit_types__]
                raise ValidationError(f"field {self.name} must be any of {type_names}")

        if self.choices and value not in self.choices:
            if value is None:
                # at this point the value can be None only if it's not required
                # thus, if it's not required, it's ok to be None even if choices are defined
                return

            raise ValidationError(f"field {self.name} must be one of {self.choices}")


class ObjectIdField(Field[ObjectId]):
    pass


class ReferenceField(ObjectIdField, Generic[TModelType]):

    on_destroy: OnDestroy

    _reference_model: TModelType

    def register_ref(self, owner: Type[object], name: str) -> None:
        if not inspect.isabstract(owner) and ABC not in owner.__bases__:
            self._reference_model.add_ref(
                ModelReference(ref_class=owner, ref_field=name, on_destroy=self.on_destroy)
            )

    def __init__(
        self,
        *,
        reference_model: TModelType,
        required: bool = False,
        rejected: bool = False,
        restricted: bool = False,
        default: str | Callable[[], str] | None = None,
        index: IndexInput | None = True,
        unique: bool = False,
        on_destroy: OnDestroy = OnDestroy.RAISE,
    ) -> None:
        if on_destroy == OnDestroy.DETACH and required:
            raise RuntimeError("cannot auto-detach a required reference")
        super().__init__(
            required=required,
            rejected=rejected,
            restricted=restricted,
            index=index,
            default=default,
            unique=unique,
        )
        self.on_destroy = on_destroy
        self._reference_model = reference_model

    async def validate(self, obj: object) -> None:
        await super().validate(obj)
        value = self.__get__(obj)
        if value is None:
            return
        ref = await self._reference_model.get(value)
        if ref is None:
            raise ValidationError(
                f"Broken reference {self.name}: no {self._reference_model.__name__} found"
            )


class SelfReferenceField(ReferenceField):

    def register_ref(self, owner: Type[object], name: str) -> None:
        if not inspect.isabstract(owner) and ABC not in owner.__bases__:
            self._reference_model = owner
            self._reference_model.add_ref(
                ModelReference(ref_class=owner, ref_field=name, on_destroy=self.on_destroy)
            )

    def __init__(
        self,
        *,
        required: bool = False,
        rejected: bool = False,
        restricted: bool = False,
        default: str | Callable[[], str] | None = None,
        index: IndexInput | None = True,
        unique: bool = False,
        on_destroy: OnDestroy = OnDestroy.RAISE,
    ) -> None:
        if on_destroy == OnDestroy.DETACH and required:
            raise RuntimeError("cannot auto-detach a required reference")
        super().__init__(
            required=required,
            rejected=rejected,
            restricted=restricted,
            index=index,
            default=default,
            unique=unique,
            reference_model=None,
        )
        self.on_destroy = on_destroy


class StringField(Field[str]):

    min_length: int | None
    max_length: int | None
    re_match: re.Pattern[str] | None
    auto_trim: bool

    __explicit_types__ = [str]

    def __init__(
        self,
        *,
        min_length: int | None = None,
        max_length: int | None = None,
        re_match: str | None = None,
        auto_trim: bool = True,
        required: bool = False,
        rejected: bool = False,
        restricted: bool = False,
        default: str | Callable[[], str] | None = None,
        index: IndexInput | None = None,
        unique: bool = False,
        choices: Iterable[str] | None = None,
    ) -> None:
        super().__init__(
            required=required,
            rejected=rejected,
            restricted=restricted,
            default=default,
            index=index,
            unique=unique,
            choices=choices,
        )

        self.auto_trim = auto_trim
        self.min_length = min_length
        self.max_length = max_length
        if re_match:
            self.re_match = re.compile(re_match)
        else:
            self.re_match = None

    def __set__(self, obj: object, value: str | None) -> None:
        # user can accidentally put something other than string to a string field
        # we must allow him to do that and only complain at validation
        # thus the hasattr check
        if hasattr(value, "strip") and self.auto_trim:
            value = value.strip()
        return super().__set__(obj, value)

    async def validate(self, obj: object) -> None:
        await super().validate(obj)
        value = self.__get__(obj)
        if value is None:
            return
        min_length = self.min_length
        if min_length is not None and len(value) < min_length:
            raise ValidationError(
                f"field {self.name} must be at least {self.min_length} characters long"
            )

        max_length = self.max_length
        if max_length is not None and len(value) > max_length:
            raise ValidationError(
                f"field {self.name} must be at most {self.max_length} characters long"
            )

        re_match = self.re_match
        if re_match is not None and not re_match.match(value):
            raise ValidationError(
                f'field {self.name} must match pattern "{re_match.pattern}"'
            )


TNumber = TypeVar("TNumber", bound=Real)


class NumberField(Field[TNumber]):

    min_value: TNumber | None
    max_value: TNumber | None

    __explicit_types__ = [int, float]

    def __init__(
        self,
        *,
        min_value: TNumber | None = None,
        max_value: TNumber | None = None,
        required: bool = False,
        rejected: bool = False,
        restricted: bool = False,
        default: TNumber | Callable[[], TNumber] | None = None,
        index: IndexInput | None = None,
        unique: bool = False,
        choices: Iterable[TNumber] | None = None,
    ) -> None:
        super().__init__(
            required=required,
            rejected=rejected,
            restricted=restricted,
            default=default,
            index=index,
            unique=unique,
            choices=choices,
        )
        self.min_value = min_value
        self.max_value = max_value

    async def validate(self, obj: object) -> None:
        await super().validate(obj)
        value = self.__get__(obj)
        if value is None:
            return
        min_value = self.min_value
        max_value = self.max_value
        if min_value is not None and value < min_value:
            raise ValidationError(f"field {self.name} must be >= {min_value}")
        if max_value is not None and value > max_value:
            raise ValidationError(f"field {self.name} must be <= {max_value}")


class IntField(NumberField[int]):
    __explicit_types__ = [int]


class FloatField(NumberField[float]):
    __explicit_types__ = [float]


class ListField(Generic[T], Field[list[T]]):

    min_length: int | None

    __explicit_types__ = [list, set, tuple]

    def __init__(
        self,
        min_length: int | None = None,
        max_length: int | None = None,
        required: bool = False,
        rejected: bool = False,
        restricted: bool = False,
        default: Callable[[], list[T]] | None = None,
        index: IndexInput | None = None,
    ) -> None:
        super().__init__(
            required=required,
            rejected=rejected,
            restricted=restricted,
            default=default,
            index=index,
            unique=False,
            choices=None,
        )
        if min_length is not None:
            assert isinstance(min_length, int)
        if max_length is not None:
            assert isinstance(max_length, int)
        self.min_length = min_length
        self.max_length = max_length

    async def validate(self, obj: object) -> None:
        await super().validate(obj)
        value = self.__get__(obj)
        if value is None:
            return
        min_length = self.min_length
        max_length = self.max_length
        if min_length is not None and len(value) < min_length:
            raise ValidationError(
                f"field {self.name} must be at least {min_length} items long"
            )

        if max_length is not None and len(value) > max_length:
            raise ValidationError(
                f"field {self.name} must be at most {self.max_length} items long"
            )


class DatetimeField(Field[datetime]):
    __explicit_types__ = [datetime]


K = TypeVar("K", bound=Hashable)
V = TypeVar("V")


class DictField(Field[dict[K, V]]):

    __explicit_types__ = [dict]

    def __init__(
        self,
        required: bool = False,
        rejected: bool = False,
        restricted: bool = False,
        default: Callable[[], dict[K, V]] | None = None,
        index: IndexInput | None = None,
    ) -> None:
        super().__init__(
            required=required,
            rejected=rejected,
            restricted=restricted,
            default=default,
            index=index,
            unique=False,
            choices=None,
        )


class BoolField(Field[bool]):
    __explicit_types__ = [bool]


@dataclass
class ComputedField:
    is_async: bool



import asyncio
import importlib
import inspect
import os.path
from argparse import ArgumentParser, Namespace
from typing import Generator, AnyStr, Type, TypeVar, List, Dict
from croydon.types import TConfigType


class Command:

    NAME: str
    HELP: str
    ASYNC_RUN: bool = True

    config: str

    def __init__(self, parser: ArgumentParser) -> None:
        self.init_argument_parser(parser)

    def run(self, args: Namespace) -> None:
        asyncio.run(self.setup(args))
        if self.ASYNC_RUN:
            asyncio.run(self.run_async())
        else:
            self.run_sync()

    async def setup(self, args: Namespace) -> None:
        pass

    def init_argument_parser(self, parser: ArgumentParser) -> None:
        pass

    def run_sync(self) -> None:
        pass

    async def run_async(self) -> None:
        pass


CC = Type[Command]
CI = TypeVar("CI", bound=Command)


def _collect_commands() -> List[CC]:

    search_module_paths = ["croydon.commands", "app.commands"]
    commands: List[CC] = []

    def module_names_in_dir(mod_dir: AnyStr) -> Generator[str, None, None]:
        for filename in os.listdir(mod_dir):
            if filename.endswith(".py"):
                filename = os.path.basename(filename)[:-3]
                yield filename

    for modpath in search_module_paths:
        try:
            base_module = importlib.import_module(modpath)
        except ModuleNotFoundError:
            continue
        for modname in module_names_in_dir(os.path.dirname(base_module.__file__)):
            module = importlib.import_module(f"{modpath}.{modname}")
            for cls in module.__dict__.values():
                if inspect.isclass(cls) and issubclass(cls, Command) and cls is not Command:
                    commands.append(cls)

    commands.sort(key=lambda cmd: cmd.NAME)

    return commands


def main():
    parser = ArgumentParser()
    subparsers = parser.add_subparsers(
        title="command",
        required=True,
        help="command to run",
        dest="command"
    )

    cmd_map: Dict[str, CI] = {}

    for cmd_class in _collect_commands():
        sp = subparsers.add_parser(cmd_class.NAME, help=cmd_class.HELP)
        cmd = cmd_class(sp)
        cmd_map[cmd.NAME] = cmd

    args = parser.parse_args()
    cmd = cmd_map.get(args.command)
    cmd.run(args)



from typing import Optional, Any
from .abc import AbstractCache


class NoCache(AbstractCache):
    NAME = "NoCache"

    async def has(self, key: str) -> bool:
        return False

    async def set(self, key: str, value: Any) -> None:
        return

    async def get(self, key: str) -> Optional[Any]:
        return None

    async def delete(self, key: str) -> bool:
        return False

    async def initialise(self) -> None:
        return



import asyncio
import functools
import binascii
import pickle
from typing import Optional, Any, List, Tuple

import aiomcache
from aiomcache import Client
from .abc import AbstractCache

MAX_RETRIES: int = 5


def pick_and_retry(func):

    @functools.wraps(func)
    async def wrapper(self, key, *args, **kwargs):
        retry = 0
        while True:
            backend, key = self._get_backend(key, retry)
            try:
                data = await func(self, backend, key, *args, **kwargs)
                return data
            except Exception:
                retry += 1
                if retry > MAX_RETRIES:
                    raise
    return wrapper


def server_hash_func(key):
    return (((binascii.crc32(key) & 0xffffffff) >> 16) & 0x7fff) or 1


class MemcachedCache(AbstractCache):
    NAME = "MemcachedCache"
    _backends: List[Client]

    def __init__(self, backends: List[str]) -> None:
        clients: List[Client] = []

        for backend in backends:
            chunks = backend.split(":")
            if len(chunks) == 1:
                port = 11211
            else:
                port = chunks[1]
            client = Client(
                chunks[0],
                port,
                pool_size=2,
                pool_minsize=None,
            )
            clients.append(client)

        self._backends = clients

    def _get_backend(self, key: str | Tuple, retry: int = 0) -> Tuple[Optional[Client], Optional[str]]:
        if not self._backends:
            return None, None

        if isinstance(key, tuple):
            serverhash, key = key
        else:
            serverhash = server_hash_func(key.encode("utf-8"))

        if retry > 0:
            serverhash = str(serverhash) + str(retry)
            serverhash = server_hash_func(serverhash.encode('utf-8'))

        server = self._backends[serverhash % len(self._backends)]
        return server, key

    async def has(self, key: str) -> bool:
        value = await self.get(key)
        return value is not None

    @pick_and_retry
    async def _set(self, backend: Client, key: str, value: Any) -> None:
        await backend.set(key.encode(), value)

    async def set(self, key: str, value: Any) -> None:
        value = pickle.dumps(value)
        try:
            await self._set(key, value)
        except aiomcache.ValidationException:
            pass

    @pick_and_retry
    async def _get(self, backend: Client, key: str) -> Optional[Any]:
        return await backend.get(key.encode("utf-8"))

    async def get(self, key: str) -> Optional[Any]:
        try:
            res = await self._get(key)
        except aiomcache.ValidationException:
            return None
        if res is not None:
            try:
                value = pickle.loads(res)
                return value
            except Exception:
                return None

    async def delete(self, key: str) -> bool:
        try:
            tasks = [backend.delete(key.encode("utf-8")) for backend in self._backends]
            results = await asyncio.gather(*tasks)
        except aiomcache.ValidationException:
            return False
        return any(results)

    async def initialise(self) -> None:
        return



from typing import Optional
from datetime import datetime, timedelta

from croydon import ctx
from croydon.models import StorableModel
from croydon.models.fields import StringField, DatetimeField, ReferenceField, OnDestroy, BoolField
from croydon.models.index import Index, IndexDirection, IndexKey
from croydon.util import now, uuid4_string

from .user import User


class Token(StorableModel):

    COLLECTION = "tokens"
    KEY_FIELD = "token"
    INDEXES = [
        Index(
            keys=[
                IndexKey("user_id", IndexDirection.ASCENDING),
                IndexKey("token_type", IndexDirection.ASCENDING),
            ]
        )
    ]

    DEFAULT_TTL = timedelta(days=90)

    token_type = StringField(default="auth", required=True, rejected=True)
    token = StringField(default=uuid4_string, required=True, rejected=True, unique=True)
    created_at = DatetimeField(default=now, required=True, rejected=True)
    updated_at = DatetimeField(default=now, required=True, rejected=True)
    expires_at = DatetimeField(default=lambda: Token.calculate_expires_at())
    description = StringField(default="")
    user_id: ReferenceField[User] = ReferenceField(
        reference_model=User,
        required=True,
        rejected=True,
        restricted=True,
        on_destroy=OnDestroy.CASCADE
    )
    auto_prolonged = BoolField(default=True)

    def touch(self):
        self.updated_at = now()

    async def user(self) -> Optional[User]:
        return await User.get(self.user_id)

    def is_expired(self) -> bool:
        return now() > self.expires_at

    @classmethod
    def calculate_expires_at(cls) -> datetime:
        return now() + cls.DEFAULT_TTL

    async def auto_prolong(self):
        if not self.auto_prolonged:
            return
        ttl = self.expires_at - now()
        if ttl < timedelta(days=7):
            self.expires_at = self.calculate_expires_at()
            await self.save()
            ctx.log.debug(f"token {self.token} auto-prolonged")



from typing import Optional
from time import time
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from fastapi import Request, Response
from croydon import ctx
from app.models.session import Session


class SessionMiddleware(BaseHTTPMiddleware):

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        session_key = ctx.cfg.session.cookie
        session_id = request.cookies.get(session_key)
        need_cookie = session_id is None
        session: Optional[Session] = None

        if session_id is not None:
            session = await Session.cache_get(session_id)

        if session is None:
            session = Session()
            need_cookie = True

        request.state.session = session
        response = await call_next(request)

        if request.state.session.is_modified():
            try:
                await request.state.session.save()
            except Exception as e:
                ctx.log.error(f"can't save session: {e}")

        if need_cookie:
            response.set_cookie(ctx.cfg.session.cookie, session.key)

        return response


class TimingMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        t1 = time()
        response = await call_next(request)
        t2 = time()
        path = request.url.path

        if request.url.query:
            path += "?" + request.url.query
        ctx.log.debug("[TIME %.3fs %s %s] ", t2-t1, request.method, path)

        return response



from typing import Optional, Any, Dict, List, Literal
from dataclasses import dataclass
from .abc import AbstractCache


CallMethod = Literal["has", "get", "set", "delete"]


@dataclass
class Call:
    args: List[Any]


class TraceCache(AbstractCache):
    """
    TraceCache is equivalent of NoCache for testing purposes.
    Besides not doing anything on its methods calls it actually traces
    all the calls and saves them for further checking / asserting
    """

    NAME = "TraceCache"

    calls: Dict[CallMethod, List[Call]]

    def __init__(self):
        super().__init__()
        self.reset()

    def reset(self):
        self.calls = {
            "has": [],
            "set": [],
            "get": [],
            "delete": [],
        }

    async def has(self, key: str) -> bool:
        self.calls["has"].append(Call(args=[key]))
        return False

    async def set(self, key: str, value: Any) -> None:
        self.calls["set"].append(Call(args=[key, value]))
        return

    async def get(self, key: str) -> Optional[Any]:
        self.calls["get"].append(Call(args=[key]))
        return None

    async def delete(self, key: str) -> bool:
        self.calls["delete"].append(Call(args=[key]))
        return False

    async def initialise(self) -> None:
        return

    def called_once(self, method: CallMethod, *args: Any) -> None:
        flt = lambda x: x.args == list(args) if args else lambda x: x
        calls = list(filter(flt, self.calls[method]))
        args_message = f" with args {args}" if args else ""
        if len(calls) < 1:
            raise AssertionError(f"method {method} was not called{args_message}")
        if len(calls) > 1:
            raise AssertionError(f"method {method} was called {len(calls)} times{args_message}")

    def called_times(self, method: CallMethod, times: int, *args: Any) -> None:
        flt = lambda x: x.args == list(args) if args else lambda x: x
        calls = list(filter(flt, self.calls[method]))
        if len(calls) != times:
            args_message = f" with args {args}" if args else ""
            raise AssertionError(
                f"method {method} was called {len(calls)} times{args_message} ({times} times expected)"
            )

    def called(self, method: CallMethod, *args: Any) -> None:
        flt = lambda x: x.args == list(args) if args else lambda x: x
        calls = list(filter(flt, self.calls[method]))
        if len(calls) == 0:
            args_message = f" with args {args}" if args else ""
            raise AssertionError(f"method {method} was not called{args_message}")

    def not_called(self, method: CallMethod, *args: Any) -> None:
        flt = lambda x: x.args == list(args) if args else lambda x: x
        calls = list(filter(flt, self.calls[method]))
        if len(calls) != 0:
            args_message = f" with args {args}" if args else ""
            raise AssertionError(f"method {method} was called{args_message} {len(calls)} times")



from typing import AsyncGenerator
from abc import abstractmethod, ABC
from .task import BaseTask
from .types import TBaseTask


class BaseQueue(ABC):

    @abstractmethod
    def enqueue(self, task: TBaseTask) -> bool:
        if not isinstance(task, BaseTask):
            raise TypeError("only instances of BaseTask subclasses are allowed")
        return False

    @abstractmethod
    def set_task_done(self, task: TBaseTask, status_message: str = "ok") -> None:
        if not isinstance(task, BaseTask):
            raise TypeError("only instances of BaseTask subclasses are allowed")

    @abstractmethod
    async def tasks(self, no_block: bool = False) -> AsyncGenerator[TBaseTask, None]: ...
